"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BelongsToRestResolver = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2023 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const constants_1 = require("../../constants");
const types_1 = require("../types");
let BelongsToRestResolver = class BelongsToRestResolver {
    constructor(context) {
        this.context = context;
        this.type = repository_1.RelationType.belongsTo;
    }
    /**
     * It takes the results of the query, finds the foreign key in each result, and then uses that key to
     * find the related record in the related service
     * @param  - `relationConfig` is the configuration object for the relation.
     * @returns A map of related records.
     */
    async resolve({ relationConfig, relationMetadata, results, scope, token, }) {
        var _a, _b;
        const relatedService = await this.getService(relationConfig);
        const keyFrom = ((_a = relationMetadata.keyFrom) !== null && _a !== void 0 ? _a : this.toPascalCase(relationMetadata.source.name) + 'Id');
        const keyTo = ((_b = relationMetadata.keyTo) !== null && _b !== void 0 ? _b : 'id');
        const filter = await this.addConditionToScope(results.map(r => r[keyFrom]), scope);
        const relatedResults = await relatedService.find(filter !== null && filter !== void 0 ? filter : {}, token, relationConfig);
        const relatedRecordMap = new Map();
        for (const item of relatedResults) {
            const key = item[keyTo];
            if (!relatedRecordMap.has(key)) {
                relatedRecordMap.set(key, item);
            }
        }
        return relatedRecordMap;
    }
    /**
     * It takes a parent object, looks up the foreign key on the parent, and then uses that foreign key to
     * look up and link the related object in the resolvedDataMap
     * @param  - `relationMetadata` is the metadata for the relation being linked.
     * @returns The parent object with the relationName property set to the
     * resolvedDataMap.get(parent[keyFrom])
     */
    async link({ relationMetadata, parent, resolvedDataMap, token, }) {
        var _a;
        const keyFrom = ((_a = relationMetadata.keyFrom) !== null && _a !== void 0 ? _a : this.toPascalCase(relationMetadata.source.name) + 'Id');
        const relationName = relationMetadata.name;
        parent[relationName] = resolvedDataMap.get(parent[keyFrom]);
        return parent;
    }
    async addConditionToScope(ids, scope) {
        const condition = {
            id: {
                inq: ids.filter(id => id), //filter out valid IDs
            },
        };
        const whereBuilder = new repository_1.WhereBuilder();
        whereBuilder.and([condition, scope === null || scope === void 0 ? void 0 : scope.where].filter(w => !!w));
        return {
            ...scope,
            where: whereBuilder.build(),
        };
    }
    toPascalCase(str) {
        return str
            .replace(/(?:^\w|[A-Z]|\b\w)/g, function (word, index) {
            return index === 0 ? word.toLowerCase() : word.toUpperCase();
        })
            .replace(/\s+/g, '');
    }
    async getService(config) {
        if ((0, types_1.isConfigWithKey)(config)) {
            return this.context.get(config.serviceKey);
        }
        else if ((0, types_1.isConfigWithModelClass)(config)) {
            return this.context.get(`services.${config.modelClass.name}Proxy`);
        }
        else {
            return this.context.get(`services.${config.serviceClass.name}`);
        }
    }
};
exports.BelongsToRestResolver = BelongsToRestResolver;
exports.BelongsToRestResolver = BelongsToRestResolver = tslib_1.__decorate([
    (0, constants_1.asRestResolver)(),
    tslib_1.__param(0, core_1.inject.context()),
    tslib_1.__metadata("design:paramtypes", [core_1.Context])
], BelongsToRestResolver);
//# sourceMappingURL=belongs-to-resolver.service.js.map