"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HasManyRestResolver = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2023 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const constants_1 = require("../../constants");
const types_1 = require("../types");
let HasManyRestResolver = class HasManyRestResolver {
    constructor(context) {
        this.context = context;
        this.type = repository_1.RelationType.hasMany;
    }
    async resolve({ relationConfig, relationMetadata, results, scope, token, }) {
        if (relationMetadata.through) {
            return this._resolveHasManyThrough({
                relationConfig,
                relationMetadata,
                results,
                scope,
                token,
            });
        }
        else {
            return this._resolveHasManyOnly({
                relationConfig,
                relationMetadata,
                results,
                scope,
                token,
            });
        }
    }
    async _resolveHasManyThrough({ relationConfig, relationMetadata, results, scope, token, }) {
        var _a, _b, _c, _d;
        if (!relationMetadata.through) {
            throw new Error('Through relation was expected');
        }
        if (!relationConfig.throughRelation) {
            throw new Error('Through relation was expected in proxy config');
        }
        const relatedService = await this.getService(relationConfig);
        const keyFromThrough = (_a = relationMetadata.through.keyFrom) !== null && _a !== void 0 ? _a : `${this.toPascalCase(relationMetadata.source.name)}Id`;
        const keyToThrough = (_b = relationMetadata.through.keyTo) !== null && _b !== void 0 ? _b : `${this.toPascalCase(relationMetadata.target.name)}Id`;
        const keyTo = ((_c = relationMetadata.keyTo) !== null && _c !== void 0 ? _c : 'id');
        const throughModelProperty = relationConfig.throughRelation;
        const idsSet = new Set();
        const throughMap = new Map();
        for (const result of results) {
            const throughData = result[throughModelProperty];
            if (throughData) {
                throughData.forEach(d => {
                    idsSet.add(d[keyToThrough]);
                    throughMap.set(d[keyToThrough], d[keyFromThrough]);
                });
            }
        }
        const filter = await this.addConditionToScope(Array.from(idsSet), keyTo, scope);
        const relatedResults = await this.findRelatedData(relatedService, filter, token, relationConfig);
        const relatedRecordMap = new Map();
        for (const item of relatedResults) {
            const keyFrom = throughMap.get(item[keyTo]);
            if (!keyFrom) {
                throw new Error(`No key found for ${item[keyTo]} in through model ${relationMetadata.through.model.name}`);
            }
            if (!relatedRecordMap.has(keyFrom)) {
                relatedRecordMap.set(keyFrom, [item]);
            }
            else {
                (_d = relatedRecordMap.get(keyFrom)) === null || _d === void 0 ? void 0 : _d.push(item);
            }
        }
        return relatedRecordMap;
    }
    async _resolveHasManyOnly({ relationConfig, relationMetadata, results, scope, token, }) {
        var _a, _b, _c;
        const relatedService = await this.getService(relationConfig);
        const keyFrom = ((_a = relationMetadata.keyFrom) !== null && _a !== void 0 ? _a : relationMetadata.source.getIdProperties()[0]);
        const keyTo = ((_b = relationMetadata.keyTo) !== null && _b !== void 0 ? _b : `${this.toPascalCase(relationMetadata.source.name)}Id`);
        const filter = await this.addConditionToScope(results.map(r => r[keyFrom]), keyTo, scope);
        const relatedResults = await this.findRelatedData(relatedService, filter, token, relationConfig);
        const relatedRecordMap = new Map();
        for (const item of relatedResults) {
            const key = item[keyTo];
            if (!relatedRecordMap.has(key)) {
                relatedRecordMap.set(key, [item]);
            }
            else {
                (_c = relatedRecordMap.get(key)) === null || _c === void 0 ? void 0 : _c.push(item);
            }
        }
        return relatedRecordMap;
    }
    async link(params) {
        var _a;
        const keyFrom = ((_a = params.relationMetadata.keyFrom) !== null && _a !== void 0 ? _a : params.relationMetadata.source.getIdProperties()[0]);
        const relationName = params.relationMetadata.name;
        params.parent[relationName] = params.resolvedDataMap.get(params.parent[keyFrom]);
        return params.parent;
    }
    async addConditionToScope(ids, field, scope) {
        const condition = {
            [field]: {
                inq: ids.filter(id => id), //filter out valid IDs
            },
        };
        const whereBuilder = new repository_1.WhereBuilder();
        whereBuilder.and([condition, scope === null || scope === void 0 ? void 0 : scope.where].filter(w => !!w));
        return {
            ...scope,
            where: whereBuilder.build(),
        };
    }
    findRelatedData(service, filter, token, config) {
        return service.find(filter, token, config);
    }
    toPascalCase(str) {
        return str
            .replace(/(?:^\w|[A-Z]|\b\w)/g, function (word, index) {
            return index === 0 ? word.toLowerCase() : word.toUpperCase();
        })
            .replace(/\s+/g, '');
    }
    async getService(config) {
        if ((0, types_1.isConfigWithKey)(config)) {
            return this.context.get(config.serviceKey);
        }
        else if ((0, types_1.isConfigWithModelClass)(config)) {
            return this.context.get(`services.${config.modelClass.name}Proxy`);
        }
        else {
            return this.context.get(`services.${config.serviceClass.name}`);
        }
    }
};
exports.HasManyRestResolver = HasManyRestResolver;
exports.HasManyRestResolver = HasManyRestResolver = tslib_1.__decorate([
    (0, constants_1.asRestResolver)(),
    tslib_1.__param(0, core_1.inject.context()),
    tslib_1.__metadata("design:paramtypes", [core_1.Context])
], HasManyRestResolver);
//# sourceMappingURL=has-many-resolver.service.js.map