"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RestServiceModifier = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2023 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const core_1 = require("@loopback/core");
const rest_1 = require("@loopback/rest");
const keys_1 = require("../keys");
let RestServiceModifier = class RestServiceModifier {
    constructor(resolverGetter, _validateToken, req) {
        this.resolverGetter = resolverGetter;
        this._validateToken = _validateToken;
        this.req = req;
        if (req === null || req === void 0 ? void 0 : req.headers.authorization) {
            this.token = req.headers.authorization;
        }
    }
    async value() {
        this.resolvers = await this.resolverGetter();
        return (service, model, config) => {
            return {
                ...service,
                create: this.create.bind(this, service),
                findById: this.findById.bind(this, service, model, config),
                find: this.find.bind(this, service, model, config),
                count: this.count.bind(this, service),
                updateById: this.updateById.bind(this, service),
                deleteById: this.deleteById.bind(this, service),
                update: this.update.bind(this, service),
                replaceById: this.replaceById.bind(this, service),
                delete: this.delete.bind(this, service),
            };
        };
    }
    create(service, data, token) {
        return service.create(this._validateToken(this, token), data);
    }
    async findById(service, model, config, id, filter, token) {
        const { restRelations, scopeMap } = this._relationsFromFilter(config, filter);
        const result = await service.findById(this._validateToken(this, token), id, JSON.stringify(filter));
        const resolvedResults = await this._resolveRelations(restRelations, scopeMap, model, [result], token);
        return resolvedResults[0];
    }
    async find(service, model, config, filter, token, inclusionConfig) {
        const { restRelations, scopeMap } = this._relationsFromFilter(config, filter);
        const passedFilter = (inclusionConfig === null || inclusionConfig === void 0 ? void 0 : inclusionConfig.disableStringify)
            ? filter
            : JSON.stringify(filter);
        const results = await service.find(this._validateToken(this, token), passedFilter);
        return this._resolveRelations(restRelations, scopeMap, model, results, token);
    }
    count(service, where, token) {
        return service.count(this._validateToken(this, token), JSON.stringify(where));
    }
    updateById(service, id, data, token) {
        return service.updateById(this._validateToken(this, token), id, data);
    }
    deleteById(service, id, token) {
        return service.deleteById(this._validateToken(this, token), id);
    }
    delete(service, where, token) {
        return service.delete(this._validateToken(this, token), where);
    }
    update(service, data, where, token) {
        return service.update(this._validateToken(this, token), data, JSON.stringify(where));
    }
    replaceById(service, id, data, token) {
        return service.replaceById(this._validateToken(this, token), id, data);
    }
    /**
     * It takes a filter and a list of relations, and returns a list of relations that are REST relations,
     * and a map of scopes for those relations
     * @param {RestRelationConfig[]} configs - RestRelationConfig[]
     * @param [filter] - The filter object that was passed to the find method.
     * @returns An object with two properties: restRelations and scopeMap.
     */
    _relationsFromFilter(configs, filter) {
        if (filter === null || filter === void 0 ? void 0 : filter.include) {
            const restRelations = [];
            const normalRelations = [];
            const scopeMap = new Map();
            filter.include.forEach(relation => {
                let name;
                if (typeof relation === 'string') {
                    name = relation;
                }
                else {
                    name = relation.relation;
                }
                const relationConfig = configs.find(config => config.name === name);
                if (relationConfig) {
                    restRelations.push(relationConfig);
                    if (typeof relation !== 'string' && relation.scope) {
                        scopeMap.set(relationConfig.name, relation.scope);
                    }
                }
                else {
                    normalRelations.push(relation);
                }
            });
            const extraIncludes = this._extraFilters(restRelations);
            this._updateFilter(extraIncludes, filter, normalRelations);
            return { restRelations, scopeMap };
        }
        return { restRelations: [], scopeMap: new Map() };
    }
    _updateFilter(extraIncludes, filter, normalRelations) {
        if (extraIncludes.length) {
            normalRelations.push(...extraIncludes);
        }
        filter.include = normalRelations;
        if (filter.include.length === 0) {
            delete filter.include;
        }
    }
    _extraFilters(configs) {
        const extraIncludes = [];
        configs.forEach(config => {
            if (config.throughRelation) {
                extraIncludes.push({
                    relation: config.throughRelation,
                });
            }
        });
        return extraIncludes;
    }
    /**
     * It takes a list of relations, a list of results, and a token, and returns a list of results with the
     * relations resolved
     * @param {RestRelationConfig[]} relations - RestRelationConfig[]
     * @param scopeMap - Map<string, Filter<S>>
     * @param model - The model class that we are querying
     * @param {T[]} results - The results of the query.
     * @param {string} token - The access token of the current user.
     */
    async _resolveRelations(relations, scopeMap, model, results, token) {
        const resolvePromises = [];
        const indexedData = [];
        for (const relationConfig of relations) {
            const relationMetadata = model.definition.relations[relationConfig.name];
            if (!relationMetadata) {
                throw rest_1.HttpErrors.BadRequest(`No metadata with name: ${relationConfig.name}`);
            }
            const resolver = this.resolvers.find(r => r.type === relationMetadata.type);
            if (resolver) {
                resolvePromises.push(resolver.resolve({
                    relationConfig,
                    relationMetadata,
                    results,
                    scope: scopeMap.get(relationConfig.name),
                    token,
                }));
                indexedData.push({
                    metadata: relationMetadata,
                    resolver,
                });
            }
            else {
                throw rest_1.HttpErrors.BadRequest(`No Resolver registered for relation type: ${relationMetadata.type}`);
            }
        }
        const resolvedData = await Promise.all(resolvePromises);
        const linkPromises = [];
        results.forEach(result => {
            const relatedData = {};
            indexedData.forEach(({ metadata, resolver }, index) => {
                linkPromises.push(resolver.link({
                    parent: result,
                    relationMetadata: metadata,
                    resolvedDataMap: resolvedData[index],
                    token,
                }));
            });
            return relatedData;
        });
        await Promise.all(linkPromises);
        return results;
    }
    _configsHaveRelations(configs, name) {
        return configs.some(config => config.name === name);
    }
};
exports.RestServiceModifier = RestServiceModifier;
exports.RestServiceModifier = RestServiceModifier = tslib_1.__decorate([
    (0, core_1.extensionPoint)(keys_1.ServiceBuilderExtensionPoint.key, {
        scope: core_1.BindingScope.REQUEST,
    }),
    tslib_1.__param(0, (0, core_1.extensions)()),
    tslib_1.__param(1, (0, core_1.inject)(keys_1.ProxyBuilderBindings.TOKEN_VALIDATOR)),
    tslib_1.__param(2, (0, core_1.inject)(rest_1.RestBindings.Http.REQUEST, { optional: true })),
    tslib_1.__metadata("design:paramtypes", [Function, Function, Object])
], RestServiceModifier);
//# sourceMappingURL=rest-service-modifier.provider.js.map